""" This module provides provides the :class:`Supplier` interface which is used to represent values that can be
calculated lazily and track provenance of such computations. """

from __future__ import annotations

import abc
from typing import Any, Callable, Generic, Iterable, Sequence, TypeVar

from kraken.core.utils import NotSet

T = TypeVar("T")
U = TypeVar("U")


class Empty(Exception):
    """Raised when a supplier cannot provide a value."""

    def __init__(self, supplier: Supplier[Any]) -> None:
        self.supplier = supplier


class Supplier(Generic[T], abc.ABC):
    """Base class for value suppliers."""

    @abc.abstractmethod
    def derived_from(self) -> Iterable[Supplier[Any]]:
        """Return an iterable that yields all suppliers that this supplier is derived from."""

    @abc.abstractmethod
    def get(self) -> T:
        """Return the value of the supplier. Depending on the implemenmtation, this may defer to other suppliers."""

    def get_or(self, fallback: U) -> T | U:
        """Return the value of the supplier, or the *fallback* value if the supplier is empty."""
        try:
            return self.get()
        except Empty:
            return fallback

    def get_or_raise(self, get_exception: Callable[[], Exception]) -> T:
        """Return the value of the supplier, or raise the exception provided by *get_exception* if empty."""
        try:
            return self.get()
        except Empty:
            raise get_exception()

    def is_empty(self) -> bool:
        """Returns `True` if the supplier is empty."""
        try:
            self.get()
        except Empty:
            return True
        else:
            return False

    def is_filled(self) -> bool:
        """Returns `True` if the supplier is not empty."""
        return not self.is_empty()

    def is_void(self) -> bool:
        return False

    def map(self, func: Callable[[T], U]) -> Supplier[U]:
        """Maps *func* over the value in the supplier."""

        this = self

        class SupplierMap(Supplier[U]):
            def derived_from(self) -> Iterable[Supplier[Any]]:
                yield this

            def get(self) -> U:
                try:
                    return func(this.get())
                except Empty:
                    raise Empty(self)

        return SupplierMap()

    def once(self) -> Supplier[T]:
        """Cache the value forever once :attr:`get` is called."""

        this = self

        class SupplierOnce(Supplier[T]):
            _value: T | NotSet = NotSet.Value
            _empty: Empty | None = None

            def derived_from(self) -> Iterable[Supplier[Any]]:
                yield this

            def get(self) -> T:
                if self._empty is not None:
                    raise Empty(self) from self._empty
                if self._value is NotSet.Value:
                    try:
                        self._value = this.get()
                    except Empty as exc:
                        self._empty = exc
                        raise Empty(self) from exc
                return self._value

        return SupplierOnce()

    def lineage(self) -> Iterable[tuple[Supplier[Any], list[Supplier[Any]]]]:
        """Iterates over all suppliers in the lineage.

        Yields:
            A supplier and the suppliers it is derived from.
        """

        stack: list[Supplier[Any]] = [self]
        while stack:
            current = stack.pop(0)
            derived_from = list(current.derived_from())
            yield current, derived_from
            stack += derived_from

    @staticmethod
    def of(value: T, derived_from: Sequence[Supplier[T]] = ()) -> Supplier[T]:
        class SupplierOf(Supplier[T]):
            def derived_from(self) -> Iterable[Supplier[Any]]:
                return derived_from

            def get(self) -> T:
                return value

        return SupplierOf()

    @staticmethod
    def of_callable(func: Callable[[], T]) -> Supplier[T]:
        class SupplierOfCallable(Supplier[T]):
            def derived_from(self) -> Iterable[Supplier[Any]]:
                return ()

            def get(self) -> T:
                return func()

        return SupplierOfCallable()

    @staticmethod
    def void(from_exc: Exception | None = None, derived_from: Sequence[Supplier[T]] = ()) -> Supplier[T]:
        """Returns a supplier that always raises :class:`Empty`."""

        class SupplierVoid(Supplier[T]):
            def derived_from(self) -> Iterable[Supplier[Any]]:
                return derived_from

            def get(self) -> T:
                raise Empty(self) from from_exc

            def is_void(self) -> bool:
                return True

        return SupplierVoid()



"""
This file serves as a proxy for Mypy type hints. Importing it in a normal Python script results in a
:class:`RuntimeError`, you can only import from this module when your code is loaded via the Pyscript loader.
"""

from kraken.core.build_context import BuildContext
from kraken.core.project import Project

__version__ = "0.2.0"

ctx: BuildContext
project: Project

raise RuntimeError(f"you cannot import from {__name__} directly; make sure your script is loaded by Kraken")



from __future__ import annotations

import dataclasses
import warnings
from typing import Any, ClassVar, Iterable, Mapping, TypeVar

import typeapi

from kraken.core.supplier import Empty, Supplier
from kraken.core.utils import NotSet

T = TypeVar("T")


def output() -> Any:
    """Assign the result of this function as a default value to a property on the class level of an :class:`Object`
    subclass to mark it as an output property. This is an alternative to using the :class:`typing.Annotated` type
    hint.

    .. code:: Example

        from kraken.core.property import Object, Property, output

        class MyObj(Object):
            a: Property[int] = output()
    """

    return Output()


class Output:
    """Used to annotate properties, to mark them as computed outputs.

    .. code:: Example

        from kraken.core.property import Object, Property, output
        from typing_extensions import Annotated

        class MyObj(Object):
            a: Annotated[Property[int], Output]
    """


@dataclasses.dataclass
class PropertyDescriptor:
    name: str
    default: Any | NotSet
    is_output: bool


class Property(Supplier[T]):
    """A property represents an input or output parameter of an :class:`Object`."""

    def __init__(self, owner: Object, name: str) -> None:
        self.owner = owner
        self.name = name
        self._value: Supplier[T] = Supplier.void()
        self._finalized = False

    def __repr__(self) -> str:
        return f"Property({self.owner}.{self.name})"

    def derived_from(self) -> Iterable[Supplier[Any]]:
        yield self._value
        yield from self._value.derived_from()

    def get(self) -> T:
        try:
            return self._value.get()
        except Empty:
            raise Empty(self)

    def set(self, value: T | Supplier[T]) -> None:
        if self._finalized:
            raise RuntimeError(f"{self} is finalized")
        if not isinstance(value, Supplier):
            value = Supplier.of(value)
        self._value = value

    def setdefault(self, value: T | Supplier[T]) -> T:
        if self._finalized:
            raise RuntimeError(f"{self} is finalized")
        if self._value.is_void():
            self.set(value)
        return self.get()

    def setfinal(self, value: T | Supplier[T]) -> None:
        self.set(value)
        self.finalize()

    def finalize(self) -> None:
        """Prevent further modification of the value in the property."""

        if not self._finalized:
            self._finalized = True
            # TODO (@NiklasRosenstein): Materializing the property value now will prevent it from being
            #       bound later in the build, e.g. if an input property takes the value of an output property.
            # derived_from = list(self.derived_from())
            # try:
            #     self._value = Supplier.of(self.get(), derived_from)
            # except Empty as exc:
            #     self._value = Supplier.void(exc, derived_from)


class Object:
    """Base class. An object's schema is declared as annotations linking to properties."""

    __schema__: ClassVar[Mapping[str, PropertyDescriptor]] = {}

    def __init_subclass__(cls) -> None:
        """Initializes the :attr:`__schema__` by introspecting the class annotations."""

        schema: dict[str, PropertyDescriptor] = {}
        base: type[Object]
        for base in cls.__bases__:
            schema.update(base.__schema__)

        for key, hint in typeapi.get_annotations(cls).items():
            hint = typeapi.of(hint)
            is_output = False

            # Unwrap annotatations.
            if isinstance(hint, typeapi.Annotated):
                is_output = Output in hint.metadata
                if any(isinstance(x, Output) for x in hint.metadata):
                    warnings.warn(
                        f"Type hint for {cls.__name__}.{key} is annotated with 'Output' instance. You should pass"
                        "the 'Output' class directly as an annotation instead. The property will still be considered "
                        "an output property.",
                        UserWarning,
                    )
                    is_output = True
                hint = hint.wrapped

            # Check if :func:`output()` was used to indicate that the property is an output property.
            if hasattr(cls, key) and isinstance(getattr(cls, key), Output):
                is_output = True
                delattr(cls, key)

            # Is the hint pointing to a Property type?
            if isinstance(hint, typeapi.Type) and hint.type == Property:
                schema[key] = PropertyDescriptor(key, NotSet.Value, is_output)

            # The attribute is annotated as an output but not actually typed as a property?
            elif is_output:
                raise RuntimeError(
                    f"Type hint for {cls.__name__}.{key} is annotated as an 'Output' property, but not actually "
                    "typed as a 'Property'."
                )

            cls.__schema__ = schema

    def __init__(self) -> None:
        """Creates :class:`Properties <Property>` for every property defined in the object's schema."""

        for key, desc in self.__schema__.items():
            setattr(self, key, Property(self, key))

    def update(self, _raise: bool = False, **property_values: Any) -> None:
        """Assign the properties in *property_values* to the properties of the given :class:`Object`. Raises a
        :class:`ValueError` if any of the properties in *property_values* are not in the objects' schema and if
        *_raise* is set to `True`. Prints a warning otherwise."""

        additional_keys = property_values.keys() - self.__schema__.keys()
        if additional_keys and _raise:
            raise ValueError(f"{type(self).__name__} does not have these properties: {additional_keys}")
        if additional_keys:
            warnings.warn(f"{type(self).__name__} does not have these properties: {additional_keys}", UserWarning)

        for key in property_values.keys() - additional_keys:
            prop: Property[Any] = getattr(self, key)
            prop.set(property_values[key])



from __future__ import annotations

import abc
import logging
from pathlib import Path
from typing import Callable, Iterable

import pkg_resources

from kraken.core.project import Project

ENTRYPOINT = "kraken.core.loader"

logger = logging.getLogger(__name__)


class BuildScriptLoader(abc.ABC):
    @abc.abstractmethod
    def detect_in_project_directory(self, project_dir: Path) -> Path | None:
        """
        Args:
            project_dir (Path): The directory to search for a build script that can be loaded.
        Returns:
            Path: The path to the file that should be loaded.
            None: If the directory does not contain a file that can be loaded.
        """

    @abc.abstractmethod
    def match_file(self, file: Path) -> bool | Path | None:
        """
        Args:
            file (Path): A file to check if it can be loaded by this loader.
        Returns:
            bool: `True` if the file can be loaded, `False` if not.
            Path: The path to the project directory belonging to the file, if the file can be loaded.
            None: If the file cannot be loaded.
        """

    @abc.abstractmethod
    def load_script(self, file: Path, project: Project) -> None:
        """
        Args:
            file (Path): The build script to load.
            project (Project): The project to load the build script into.
        """


def get_loader_implementations() -> Iterable[BuildScriptLoader]:
    """Iterate over the registered loader implementations."""

    for ep in pkg_resources.iter_entry_points(ENTRYPOINT):
        try:
            factory: Callable[[], BuildScriptLoader] = ep.load()
            yield factory()
        except Exception:
            logger.exception("an unhandled exception occurred while fetching BuildScriptLoader implementation: %s", ep)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'core': 'src/core',
 'core.loader': 'src/core/loader',
 'testing': 'src/testing'}

packages = \
['api', 'core', 'core.loader', 'testing']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.6.0,<3.0.0', 'setuptools>=33.1.0', 'typeapi>=0.2.1,<0.3.0']

extras_require = \
{'testing': ['pytest>=6.0.0']}

entry_points = \
{'kraken.core.loader': ['pyscript = '
                        'kraken.core.loader.pyscript:PyscriptLoader']}

setup_kwargs = {
    'name': 'kraken-core',
    'version': '0.2.0',
    'description': '',
    'long_description': '# kraken-core\n\nThe `kraken.core` package provides the primitives of describing a build and deriving build tasks.\n\nAside from the `kraken.core` package, this package also provides the `kraken.api` module that is\nused only at runtime by Kraken build scripts and the `kraken.testing` module for Pytest fixtures.\n\n## How does it work?\n\nKraken uses **tasks** to describe units of work that can be chained and establish dependencies between each other.\nEach task has a **schema** that defines its input and output properties. When an output property is linked to the\ninput property of another task, this established as dependency between the tasks.\n\n```py\nfrom kraken.std.docker_gen import generate_dockerfile\nfrom kraken.std.docker_build import build_docker_image\ndockerfile = generate_dockerfile(source_file="Dockerfile.yml")\nbuild_docker_image(dockerfile=dockerfile.path, tags=["example:latest"], load=True)\n```\n\nThis populates the project with two **tasks** and connects the computed output property of one to the other,\nallowing the tasks that will run for `build_docker_image()` to pick up the dynamically generated Dockerfile that\nis written into a location in the build directory by the `generate_dockerfile()` task.\n\n<p align="center"><img src="assets/graph.png" height="225px"></p>\n\n## Core API\n\nKraken **tasks** are described with a schema. Each schema field has a type and may be an input or output parameter.\nOutput parameters are only available once a resource is executed; Kraken will that a proper execution order is\nestablished such that output properties are hydrated before another resource tries to access them as an input.\n\n```py\nfrom kraken.core.task import Context, Task, Property, Output, task_factory\nfrom typing_extensions import Annotated\n\nclass GenerateDockerfileTask(Task):\n    source_file: Property[str]\n    path: Annotated[Property[str], Output]\n\n    def execute(self, ctx: Context) -> None:\n        path = Path(self.path.setdefault(str(ctx.build_directory / "Dockerfile")))\n        path.write_text(render_dockerfile(Path(self.source_file.get()).read_text()))\n\ngenerate_dockerfile = task_factory(GenerateDockerfileTask)\n```\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
